"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.assertStructOrLinkTypeNode = exports.isStructOrLinkTypeNode = exports.assertTypeNode = exports.isTypeNode = exports.createTypeNodeFromIdl = void 0;
const idl_1 = require("../idl");
const shared_1 = require("../shared");
const ArrayTypeNode_1 = require("./ArrayTypeNode");
const BoolTypeNode_1 = require("./BoolTypeNode");
const BytesTypeNode_1 = require("./BytesTypeNode");
const EnumTypeNode_1 = require("./EnumTypeNode");
const LinkTypeNode_1 = require("./LinkTypeNode");
const MapTypeNode_1 = require("./MapTypeNode");
const NumberTypeNode_1 = require("./NumberTypeNode");
const OptionTypeNode_1 = require("./OptionTypeNode");
const PublicKeyTypeNode_1 = require("./PublicKeyTypeNode");
const SetTypeNode_1 = require("./SetTypeNode");
const StringTypeNode_1 = require("./StringTypeNode");
const StructTypeNode_1 = require("./StructTypeNode");
const TupleTypeNode_1 = require("./TupleTypeNode");
const TYPE_NODE_CLASSES = [
    'arrayTypeNode',
    'boolTypeNode',
    'bytesTypeNode',
    'linkTypeNode',
    'enumTypeNode',
    'mapTypeNode',
    'numberTypeNode',
    'numberWrapperTypeNode',
    'optionTypeNode',
    'publicKeyTypeNode',
    'setTypeNode',
    'stringTypeNode',
    'structTypeNode',
    'tupleTypeNode',
];
function isArrayOfSize(array, size) {
    return Array.isArray(array) && array.length === size;
}
const createTypeNodeFromIdl = (idlType) => {
    // Leaf.
    if (typeof idlType === 'string' && idl_1.IDL_TYPE_LEAVES.includes(idlType)) {
        if (idlType === 'bool')
            return (0, BoolTypeNode_1.boolTypeNode)();
        if (idlType === 'string')
            return (0, StringTypeNode_1.stringTypeNode)();
        if (idlType === 'publicKey')
            return (0, PublicKeyTypeNode_1.publicKeyTypeNode)();
        if (idlType === 'bytes')
            return (0, BytesTypeNode_1.bytesTypeNode)((0, shared_1.prefixedSize)());
        return (0, NumberTypeNode_1.numberTypeNode)(idlType);
    }
    // Ensure eveything else is an object.
    if (typeof idlType !== 'object') {
        throw new Error(`TypeNode: Unsupported type ${JSON.stringify(idlType)}`);
    }
    // Array.
    if ('array' in idlType && isArrayOfSize(idlType.array, 2)) {
        return (0, ArrayTypeNode_1.arrayTypeNodeFromIdl)(idlType);
    }
    // Vec.
    if ('vec' in idlType) {
        return (0, ArrayTypeNode_1.arrayTypeNodeFromIdl)(idlType);
    }
    // Defined link.
    if ('defined' in idlType && typeof idlType.defined === 'string') {
        return (0, LinkTypeNode_1.linkTypeNode)(idlType.defined);
    }
    // Enum.
    if ('kind' in idlType && idlType.kind === 'enum' && 'variants' in idlType) {
        return (0, EnumTypeNode_1.enumTypeNodeFromIdl)(idlType);
    }
    // Map.
    if (('hashMap' in idlType && isArrayOfSize(idlType.hashMap, 2)) ||
        ('bTreeMap' in idlType && isArrayOfSize(idlType.bTreeMap, 2))) {
        return (0, MapTypeNode_1.mapTypeNodeFromIdl)(idlType);
    }
    // Option.
    if ('option' in idlType || 'coption' in idlType) {
        return (0, OptionTypeNode_1.optionTypeNodeFromIdl)(idlType);
    }
    // Set.
    if ('hashSet' in idlType || 'bTreeSet' in idlType) {
        return (0, SetTypeNode_1.setTypeNodeFromIdl)(idlType);
    }
    // Struct.
    if ('kind' in idlType && idlType.kind === 'struct') {
        return (0, StructTypeNode_1.structTypeNodeFromIdl)(idlType);
    }
    // Tuple.
    if ('tuple' in idlType && Array.isArray(idlType.tuple)) {
        return (0, TupleTypeNode_1.tupleTypeNodeFromIdl)(idlType);
    }
    // Throw an error for unsupported types.
    throw new Error(`TypeNode: Unsupported type ${JSON.stringify(idlType)}`);
};
exports.createTypeNodeFromIdl = createTypeNodeFromIdl;
function isTypeNode(node) {
    return !!node && TYPE_NODE_CLASSES.includes(node.kind);
}
exports.isTypeNode = isTypeNode;
function assertTypeNode(node) {
    var _a;
    if (!isTypeNode(node)) {
        throw new Error(`Expected typeNode, got ${(_a = node === null || node === void 0 ? void 0 : node.kind) !== null && _a !== void 0 ? _a : 'null'}.`);
    }
}
exports.assertTypeNode = assertTypeNode;
function isStructOrLinkTypeNode(node) {
    return (!!node && (node.kind === 'structTypeNode' || node.kind === 'linkTypeNode'));
}
exports.isStructOrLinkTypeNode = isStructOrLinkTypeNode;
function assertStructOrLinkTypeNode(node) {
    var _a;
    if (!isStructOrLinkTypeNode(node)) {
        throw new Error(`Expected structTypeNode | LinkTypeNode, got ${(_a = node === null || node === void 0 ? void 0 : node.kind) !== null && _a !== void 0 ? _a : 'null'}.`);
    }
}
exports.assertStructOrLinkTypeNode = assertStructOrLinkTypeNode;
//# sourceMappingURL=TypeNode.js.map